/*
 * kerninst.h - declarations for functions & variable used by kerninst driver
 */

#include <linux/fs.h> // struct inode
#include <linux/module.h> // module macros
#include <linux/spinlock.h> // spinlock_t
#include <asm/ptrace.h> // struct pt_regs

#include "all_vtimers.h"
#include "hash_table.h"
#include "loggedAllocRequests.h"
#include "loggedWrites.h"
#include "stacklist.h"
#include "cswitch_inout.h"

// NOTE: This file should only be included by kerninst.c due to the
//       macro and static variable definitions

MODULE_AUTHOR ("Michael J. Brim, mjbrim@cs.wisc.edu");
MODULE_DESCRIPTION ("kerninst driver v2.1.1 for Linux");
MODULE_LICENSE ("Paradyn/Kerninst");

/*--------------- Module File Operations ---------------*/
int kerninst_ioctl(struct inode *, struct file *, u_int, u_long);
int kerninst_open(struct inode *, struct file *);
int kerninst_release(struct inode *, struct file *);

static struct file_operations kerninst_file_ops = {
  owner: THIS_MODULE,
  ioctl: kerninst_ioctl,
  open: kerninst_open,
  release: kerninst_release,
  /* all other fields null */
};

/*--------------- Module Global Variables ---------------*/
static int kerninst_major = 0;
static int in_use = 0;
static spinlock_t kerninst_use_lock = SPIN_LOCK_UNLOCKED;

/*--------------- Context-Switch Handling ---------------*/

/* spinlock for use in treating cswitch code as critical section */
spinlock_t kerninst_cswitch_lock = SPIN_LOCK_UNLOCKED;

/* vector of all vtimer addresses */
struct all_vtimers av_vtimers;

/* stacklist heap of vector-of-vtimers */
struct stacklist sl_heap;

/* hash table of thread/process to vtimer list */
struct hash_table ht_vtimers;

/* count of the number of times we fail to save a vtimer on
   context-switch out due to the stacklist being full */
unsigned number_unsaved_timers;

/*--------------- Undoable Memory Operations ---------------*/

/* maintained on a per-connection basis */
static loggedWrites *global_loggedWrites = NULL;
static loggedAllocRequests *global_loggedAllocRequests = NULL;

/*--------------- Miscellaneous ---------------*/
int kerninst_call_once(void);

uint32_t kerninst_debug_bits; /* bitfield to control debug output */
uint32_t kerninst_enable_bits; /* bitfield to enable/disable functionality */

/*--------------- Platform-Specific ---------------*/
#ifdef i386_unknown_linux2_4
#include "kerninst-x86.h"
#elif defined(ppc64_unknown_linux2_4)
#include "kerninst-ppc64.h"
#else
#error "a valid platform must be defined"
#endif

/*--------------- END OF KERNINST.H ---------------*/
